=head1 NAME

Text::Gr2::CharInfo - Character information

=head1 METHODS

=head2 break_weight

How preferable it is to break at this character.

=head2 unicode_char

The unicode value of this character.

=cut

1;
